from typing import Any
import threading

from bussdcc.clock import Clock

from .event import Event
from .protocol import EventHandler, SubscriptionProtocol, EventEngineProtocol


class Subscription:
    def __init__(self, engine: EventEngineProtocol, fn: EventHandler):
        self._engine = engine
        self._fn = fn
        self._active = True

    def cancel(self) -> None:
        if self._active:
            self._engine.unsubscribe(self._fn)
            self._active = False


class EventEngine:
    def __init__(self, clock: Clock) -> None:
        self.clock = clock
        self._lock = threading.RLock()
        self._subs: list[EventHandler] = []

    def subscribe(self, fn: EventHandler) -> SubscriptionProtocol:
        with self._lock:
            self._subs.append(fn)
        return Subscription(self, fn)

    def unsubscribe(self, fn: EventHandler) -> None:
        with self._lock:
            if fn in self._subs:
                self._subs.remove(fn)

    def emit(self, name: str, **data: Any) -> Event:
        evt = Event(
            time=self.clock.now_utc(),
            name=name,
            data=data,
        )

        with self._lock:
            subs = list(self._subs)

        errors = []

        for fn in subs:
            try:
                fn(evt)
            except Exception as e:
                errors.append(
                    {
                        "event": evt.name,
                        "subscriber": repr(fn),
                        "error": repr(e),
                    }
                )

        if name != "event.subscriber_error":
            for error in errors:
                self.emit("event.subscriber_error", **error)

        return evt



from bussdcc.device import DeviceProtocol
from bussdcc.context import ContextProtocol


class Device(DeviceProtocol):
    kind: str = "device"
    id: str
    ctx: ContextProtocol | None
    online: bool

    def __init__(self, *, id: str) -> None:
        self.id = id
        self.ctx = None
        self.online = False

    def attach(self, ctx: ContextProtocol) -> None:
        self.ctx = ctx
        try:
            self.connect()
        except Exception as e:
            if self.ctx:
                self.ctx.events.emit(
                    "device.failed", device=self.id, kind=self.kind, error=repr(e)
                )
            raise
        self.online = True
        if self.ctx:
            self.ctx.events.emit("device.attached", device=self.id, kind=self.kind)

    def detach(self) -> None:
        try:
            self.disconnect()
        except Exception as e:
            if self.ctx:
                self.ctx.events.emit(
                    "device.failed", device=self.id, kind=self.kind, error=repr(e)
                )
        finally:
            self.online = False
            if self.ctx:
                self.ctx.events.emit("device.detached", device=self.id, kind=self.kind)

    def connect(self) -> None:
        """Open hardware / resources"""
        pass

    def disconnect(self) -> None:
        """Close hardware / resources"""
        pass



from typing import Protocol

from bussdcc.context import ContextProtocol
from bussdcc.event.event import Event


class PolicyProtocol(Protocol):
    name: str

    def evaluate(self, ctx: ContextProtocol, evt: Event | None = None) -> bool: ...



from typing import Protocol

from bussdcc.context import ContextProtocol


class ServiceProtocol(Protocol):
    name: str

    def start(self, ctx: ContextProtocol) -> None: ...
    def tick(self, ctx: ContextProtocol) -> None: ...
    def stop(self, ctx: ContextProtocol) -> None: ...



from .engine import Subscription, EventEngine
from .protocol import EventSink, EventHandler, EventEngineProtocol, SubscriptionProtocol
from .event import Event

__all__ = [
    "Event",
    "EventSink",
    "EventHandler",
    "EventEngine",
    "EventEngineProtocol",
    "Subscription",
    "SubscriptionProtocol",
]



import threading
from typing import Optional

from bussdcc.clock import Clock
from bussdcc.event import EventEngineProtocol
from bussdcc.state import StateEngineProtocol

from .runtime import Runtime


class ThreadedRuntime(Runtime):
    """
    Runtime variant that owns a background execution thread.

    - boot() and shutdown() remain synchronous
    - run() blocks the caller until shutdown
    - No OS signals required
    """

    def __init__(
        self,
        *,
        clock: Optional[Clock] = None,
        events: EventEngineProtocol | None = None,
        state: StateEngineProtocol | None = None,
    ):
        super().__init__(clock=clock, events=events, state=state)
        self._stop_event = threading.Event()

    def run(self) -> None:
        """
        Boot the runtime and block until shutdown is complete.
        """
        self._stop_event.clear()

        with self:
            self._run_thread = threading.current_thread()
            self._stop_event.wait()

    def shutdown(self, reason: Optional[str] = None) -> None:
        """
        Signal the runtime thread to exit, then perform shutdown.
        """
        if not self.booted:
            return

        try:
            super().shutdown(reason=reason)
        finally:
            self._stop_event.set()



from typing import Protocol, Callable, Any

from bussdcc.context import ContextProtocol

from .event import Event

EventHandler = Callable[[Event], None]


class SubscriptionProtocol(Protocol):
    def cancel(self) -> None: ...


class EventSink(Protocol):
    def start(self, ctx: ContextProtocol) -> None: ...
    def stop(self) -> None: ...
    def handle(self, evt: Event) -> None: ...


class EventEngineProtocol(Protocol):
    def emit(self, name: str, **data: Any) -> Event: ...
    def subscribe(self, handler: EventHandler) -> SubscriptionProtocol: ...
    def unsubscribe(self, fn: EventHandler) -> None: ...



from typing import Protocol, TYPE_CHECKING

from bussdcc.clock import Clock
from bussdcc.event.protocol import EventEngineProtocol
from bussdcc.state.protocol import StateEngineProtocol

if TYPE_CHECKING:
    from bussdcc.runtime.protocol import RuntimeProtocol


class ContextProtocol(Protocol):
    clock: Clock
    runtime: "RuntimeProtocol"
    events: EventEngineProtocol
    state: StateEngineProtocol



from typing import Protocol, Any


class StateEngineProtocol(Protocol):
    """
    Protocol for a hierarchical, thread-safe state store.

    Provides dot-separated path access.
    """

    def set(self, path: str, value: Any) -> None:
        """
        Set a value in the state at the given dot-separated path.

        Args:
            path: Dot-separated key path (e.g., "system.clock.uptime")
            value: Value to store
        """
        ...

    def get(self, path: str, default: Any = None) -> Any:
        """
        Retrieve a value from the state at the given dot-separated path.

        Args:
            path: Dot-separated key path
            default: Value to return if path is not found

        Returns:
            The stored value or `default` if not found
        """
        ...



from typing import Protocol

from bussdcc.context import ContextProtocol


class DeviceProtocol(Protocol):
    """
    Protocol for a device.
    Devices can attach to a context, manage their lifecycle, and report status.
    """

    kind: str
    id: str
    ctx: ContextProtocol | None
    online: bool

    def attach(self, ctx: ContextProtocol) -> None: ...
    def detach(self) -> None: ...
    def connect(self) -> None: ...
    def disconnect(self) -> None: ...



from types import FrameType
import signal
from typing import Dict, Callable

from .threaded import ThreadedRuntime


class SignalRuntime(ThreadedRuntime):
    """
    Threaded runtime supervised by POSIX signals.
    """

    SIGNAL_ACTIONS: Dict[int, str] = {
        int(signal.SIGINT): "shutdown",
        int(signal.SIGTERM): "shutdown",
        int(signal.SIGHUP): "reload",
        int(signal.SIGUSR1): "user1",
        int(signal.SIGUSR2): "user2",
    }

    def _signal_handler(self, signum: int, frame: FrameType | None) -> None:
        action = self.SIGNAL_ACTIONS.get(signum)

        if action == "shutdown":
            # Thread-safe: shutdown() only sets state and events
            self.shutdown(signal.strsignal(signum))
        elif action == "reload":
            self.ctx.events.emit("system.reload")
        elif action:
            self.ctx.events.emit("system.signal", signal=signum, action=action)

    def _on_boot(self) -> None:
        super()._on_boot()

        self._prev_handlers = {}
        for signum in self.SIGNAL_ACTIONS:
            self._prev_handlers[signum] = signal.getsignal(signum)
            signal.signal(signum, self._signal_handler)

    def _on_shutdown(self, reason: str | None = None) -> None:
        try:
            for signum, handler in self._prev_handlers.items():
                signal.signal(signum, handler)
        finally:
            super()._on_shutdown(reason)



from datetime import datetime, timezone
import time

from .protocol import Clock


class SystemClock(Clock):
    def __init__(self) -> None:
        self._origin = time.monotonic()

    def now_utc(self) -> datetime:
        return datetime.now(timezone.utc)

    def monotonic(self) -> float:
        return time.monotonic()

    def uptime(self) -> float:
        return time.monotonic() - self._origin

    def sleep(self, seconds: float) -> None:
        time.sleep(seconds)



from typing import Protocol
from datetime import datetime


class Clock(Protocol):
    def sleep(self, seconds: float) -> None: ...
    def monotonic(self) -> float: ...
    def now_utc(self) -> datetime: ...



from typing import Any
import threading

from .protocol import StateEngineProtocol


class StateEngine(StateEngineProtocol):
    """
    Thread-safe hierarchical state storage.

    Supports dot-separated paths:
        se = StateEngine()
        se.set("system.clock.uptime", 42)
        se.get("system.clock.uptime") -> 42
    """

    def __init__(self) -> None:
        self._lock = threading.RLock()
        self._state: dict[str, Any] = {}

    def set(self, path: str, value: Any) -> None:
        with self._lock:
            keys = path.split(".")
            d = self._state
            for k in keys[:-1]:
                d = d.setdefault(k, {})
            d[keys[-1]] = value

    def get(self, path: str, default: Any = None) -> Any:
        with self._lock:
            keys = path.split(".")
            d = self._state
            for k in keys:
                if not isinstance(d, dict) or k not in d:
                    return default
                d = d[k]
            return d



from typing import Dict
import threading
import traceback

from bussdcc.context import ContextProtocol

from .protocol import ServiceProtocol


class ServiceSupervisor:
    """
    Orchestrates services: start, tick, stop, and restart.
    """

    def __init__(self, ctx: ContextProtocol):
        self.ctx = ctx
        self._services: Dict[str, ServiceProtocol] = {}
        self._threads: Dict[str, threading.Thread] = {}
        self._stop_flag = threading.Event()

    def register(self, service: ServiceProtocol) -> None:
        self._services[service.name] = service

    def start_all(self) -> None:
        self._stop_flag.clear()
        for service in self._services.values():
            self._start_service(service)

    def stop_all(self) -> None:
        self._stop_flag.set()
        for service in self._services.values():
            service.stop(self.ctx)

        # Wait for threads to exit
        for t in self._threads.values():
            t.join()
        self._threads.clear()

    def _start_service(self, service: ServiceProtocol) -> None:
        def runner() -> None:
            while not self._stop_flag.is_set():
                try:
                    service.start(self.ctx)
                    while (
                        getattr(service, "enabled", True)
                        and not self._stop_flag.is_set()
                    ):
                        service.tick(self.ctx)
                        self.ctx.clock.sleep(getattr(service, "interval", 1.0))
                except Exception as e:
                    self.ctx.events.emit(
                        "service.error",
                        service=service.name,
                        error=repr(e),
                        traceback=traceback.format_exc(),
                    )
                    if getattr(service, "critical", False):
                        self.ctx.events.emit(
                            "service.critical_failure",
                            service=service.name,
                            error=repr(e),
                        )
                        # Critical failure halts supervisor
                        self._stop_flag.set()
                        break

                    if getattr(service, "restart", True):
                        self.ctx.events.emit(
                            "service.restart",
                            service=service.name,
                        )
                        continue  # restart the loop
                    else:
                        break
                finally:
                    service.stop(self.ctx)
                    self.ctx.events.emit("service.stopped", service=service.name)

        t = threading.Thread(target=runner, name=f"service:{service.name}", daemon=True)
        self._threads[service.name] = t
        t.start()
        self.ctx.events.emit("service.started", service=service.name)



from abc import abstractmethod

from bussdcc.event.event import Event
from bussdcc.context import ContextProtocol

from .protocol import ProcessProtocol


class Process(ProcessProtocol):
    name = "unnamed"
    ctx: ContextProtocol

    def _on_event(self, evt: Event) -> None:
        try:
            self.on_event(self.ctx, evt)
        except Exception as e:
            self.ctx.events.emit(
                "process.error",
                process=self.name,
                error=repr(e),
                evt=evt.name,
            )

    def attach(self, ctx: ContextProtocol) -> None:
        self.ctx = ctx
        self._sub = ctx.events.subscribe(self._on_event)
        self.on_start(ctx)
        self.ctx.events.emit("process.started", process=self.name)

    def detach(self) -> None:
        try:
            self._sub.cancel()
            self.on_stop(self.ctx)
        finally:
            self.ctx.events.emit("process.stopped", process=self.name)

    def on_start(self, ctx: ContextProtocol) -> None:
        pass

    def on_stop(self, ctx: ContextProtocol) -> None:
        pass

    @abstractmethod
    def on_event(self, ctx: ContextProtocol, evt: Event) -> None:
        pass



from bussdcc.context.protocol import ContextProtocol

from .protocol import ServiceProtocol


class Service(ServiceProtocol):
    name = "unnamed"

    # execution model
    interval = 1.0  # seconds between ticks
    enabled = True  # start at boot
    restart = True  # restart if it crashes
    critical = False  # if True, system may halt on failure

    def start(self, ctx: ContextProtocol) -> None:
        """Called once when the service starts"""
        pass

    def tick(self, ctx: ContextProtocol) -> None:
        """Called repeatedly while the service is running"""
        pass

    def stop(self, ctx: ContextProtocol) -> None:
        """Called once when the service is stopping"""
        pass



from typing import Protocol

from bussdcc.context import ContextProtocol
from bussdcc.event.event import Event


class ProcessProtocol(Protocol):
    name: str

    def attach(self, ctx: ContextProtocol) -> None: ...
    def detach(self) -> None: ...
    def on_start(self, ctx: ContextProtocol) -> None: ...
    def on_stop(self, ctx: ContextProtocol) -> None: ...
    def on_event(self, ctx: ContextProtocol, evt: Event) -> None: ...



from typing import Optional, Dict, Self, Type, Literal
from types import TracebackType

from bussdcc.context import Context, ContextProtocol
from bussdcc.clock import Clock, SystemClock
from bussdcc.device import DeviceProtocol
from bussdcc.event import Event, EventSink, EventEngine, EventEngineProtocol
from bussdcc.state import StateEngine, StateEngineProtocol
from bussdcc.service import ServiceProtocol, ServiceSupervisor
from bussdcc.process import ProcessProtocol
from bussdcc.version import get_version

from .protocol import RuntimeProtocol


class Runtime(RuntimeProtocol):
    def __init__(
        self,
        *,
        clock: Optional[Clock] = None,
        events: EventEngineProtocol | None = None,
        state: StateEngineProtocol | None = None,
    ):
        self.clock: Clock = clock or SystemClock()
        self.events: EventEngineProtocol = events or EventEngine(clock=self.clock)
        self.state: StateEngineProtocol = state or StateEngine()
        self._sinks: list[EventSink] = []
        self._devices: Dict[str, DeviceProtocol] = {}
        self._services: Dict[str, ServiceProtocol] = {}
        self._processes: Dict[str, ProcessProtocol] = {}
        self._interfaces: Dict[str, ProcessProtocol] = {}
        # type-safe context using RuntimeProtocol
        self.ctx: ContextProtocol = Context(
            clock=self.clock, runtime=self, events=self.events, state=self.state
        )
        self._booted: bool = False
        self.version: str = get_version()
        self._service_supervisor: ServiceSupervisor | None = None

    def _on_boot(self) -> None:
        """Hook for subclasses to initialize resources."""
        return None

    def _on_shutdown(self, reason: Optional[str] = None) -> None:
        """Hook for subclasses to release resources."""
        return None

    def _dispatch(self, evt: Event) -> None:
        for sink in self._sinks:
            try:
                sink.handle(evt)
            except Exception as e:
                if evt.name != "system.sink.failure":
                    self.events.emit(
                        "system.sink.failure", sink=type(sink).__name__, error=repr(e)
                    )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} booted={self._booted}>"

    def __enter__(self) -> Self:
        self.boot()

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        reason = None
        if exc_type is not None:
            reason = f"{exc_type.__name__}: {exc_val}"
        self.shutdown(reason=reason)

        # Returning False allows exceptions to propagate normally
        return False

    @property
    def booted(self) -> bool:
        return self._booted

    def add_sink(self, sink: EventSink) -> None:
        if self._booted:
            raise RuntimeError("Cannot add sinks after boot")
        self._sinks.append(sink)

    def register_service(self, service: ServiceProtocol) -> None:
        if self._booted:
            raise RuntimeError("Cannot register services after boot")
        self._services[service.name] = service

    def register_process(self, process: ProcessProtocol) -> None:
        if self._booted:
            raise RuntimeError("Cannot register processes after boot")
        self._processes[process.name] = process

    def register_interface(self, interface: ProcessProtocol) -> None:
        if self._booted:
            raise RuntimeError("Cannot register interfaces after boot")
        self._interfaces[interface.name] = interface

    def attach_device(self, device: DeviceProtocol) -> None:
        """
        Attach a device to the runtime.

        - If called before boot(), the device is staged and attached during boot.
        - If called after boot(), the device is attached immediately.
        """
        if device.id in self._devices:
            raise ValueError(f"Device {device.id} already attached")

        if not self._booted:
            self._devices[device.id] = device
            return

        device.attach(self.ctx)
        self._devices[device.id] = device

    def detach_device(self, id: str) -> None:
        device = self._devices.pop(id, None)
        if not device:
            return
        device.detach()

    def get_device(self, id: str) -> DeviceProtocol | None:
        return self._devices.get(id)

    def list_devices(self, *, kind: str | None = None) -> list[DeviceProtocol]:
        if kind is None:
            return list(self._devices.values())
        return [d for d in self._devices.values() if d.kind == kind]

    def boot(self) -> None:
        if self._booted:
            return

        self._sub = self.events.subscribe(self._dispatch)

        for sink in self._sinks:
            sink.start(self.ctx)

        self.ctx.events.emit("system.booting", version=self.version)

        self._on_boot()

        # Boot all registered devices
        for device in self._devices.values():
            device.attach(self.ctx)

        # Attach processes
        for process in self._processes.values():
            process.attach(self.ctx)

        # Attach interfaces
        for interface in self._interfaces.values():
            interface.attach(self.ctx)

        self._booted = True
        self.ctx.events.emit("system.booted", version=self.version)

        # Start services under supervisor
        self._service_supervisor = ServiceSupervisor(self.ctx)
        for service in self._services.values():
            self._service_supervisor.register(service)
        self._service_supervisor.start_all()

    def shutdown(self, reason: Optional[str] = None) -> None:
        if not self._booted:
            return  # Be idempotent on shutdown to avoid exit crashes

        try:
            self.ctx.events.emit("system.shutting_down", reason=reason)

            # Stop services first
            if self._service_supervisor:
                self._service_supervisor.stop_all()

            # Detach interfaces
            for interface in self._interfaces.values():
                interface.detach()

            # Detach processes
            for process in self._processes.values():
                process.detach()

            # Detach devices in reverse order
            for device in reversed(list(self._devices.values())):
                device.detach()

            self._on_shutdown(reason)

            self.ctx.events.emit("system.shutdown", version=self.version)

            self._sub.cancel()

            for sink in reversed(self._sinks):
                sink.stop()
        except Exception:
            raise
        finally:
            # Ensure state is reset even if on_shutdown fails
            self._booted = False



from bussdcc.clock import Clock
from bussdcc.runtime.protocol import RuntimeProtocol
from bussdcc.event.protocol import EventEngineProtocol
from bussdcc.state.protocol import StateEngineProtocol

from .protocol import ContextProtocol


class Context(ContextProtocol):
    def __init__(
        self,
        clock: Clock,
        runtime: RuntimeProtocol,
        events: EventEngineProtocol,
        state: StateEngineProtocol,
    ):
        self.clock: Clock = clock
        self.runtime: RuntimeProtocol = runtime
        self.events = events
        self.state = state



from bussdcc.event.event import Event
from bussdcc.context import ContextProtocol

from .protocol import PolicyProtocol


class Policy(PolicyProtocol):
    name = "allow_all"

    def evaluate(self, ctx: ContextProtocol, evt: Event | None = None) -> bool:
        return True



from typing import Protocol

from bussdcc.device import DeviceProtocol
from bussdcc.context import ContextProtocol


class RuntimeProtocol(Protocol):
    ctx: ContextProtocol

    @property
    def booted(self) -> bool: ...

    def attach_device(self, device: DeviceProtocol) -> None: ...
    def detach_device(self, id: str) -> None: ...
    def get_device(self, id: str) -> DeviceProtocol | None: ...
    def list_devices(self, *, kind: str | None = None) -> list[DeviceProtocol]: ...



from .runtime import Runtime
from .protocol import RuntimeProtocol
from .threaded import ThreadedRuntime
from .signal import SignalRuntime

__all__ = [
    "RuntimeProtocol",
    "Runtime",
    "ThreadedRuntime",
    "SignalRuntime",
]
